"{ Encoding: utf8 }"

"{ Package: 'stx:libbasic2' }"

"{ NameSpace: Smalltalk }"

LimitedPrecisionReal variableByteSubclass:#IEEEFloat
	instanceVariableNames:'exponentSize'
	classVariableNames:''
	poolDictionaries:''
	category:'Magnitude-Numbers'
!

!IEEEFloat class methodsFor:'documentation'!

documentation
"
    Unfinished, ongoing work

    soft float emulation for arbitrary IEEE float formats.
    This is very very slow and should only be used when importing 
    funny sized floating point numbers (such as float24 or float8) from
    external sources, or to simulate computations on otherwise unsupported floating pnt numbers.
"
! !

!IEEEFloat class methodsFor:'instance creation'!

fromFloat:aFloat
    |nB f|

    nB := aFloat byteSize.
    f := self basicNew:nB.
    1 to:nB do:[:i |
        f basicAt:i put:(aFloat byteAt:i)
    ].
    f exponentSize:(aFloat numBitsInExponent).
    ^ f

    "
     self fromFloat:1.0
     self fromFloat:2.0
     self fromFloat:1.0 asShortFloat
     self fromFloat:2.0 asShortFloat
     self fromFloat:1.0 asLongFloat
     self fromFloat:2.0 asLongFloat
    "
!

size:numBits exponentSize:exponentSize
    ^ (self basicNew:(numBits // 8)) exponentSize:exponentSize

    "
     self size:256 exponentSize:19
     self size:16 exponentSize:5
     self size:(1.0 basicSize * 8) exponentSize:(1.0 numBitsInExponent)
    "
!

size:numBits exponentSize:exponentSize fraction:anInteger exponent:exp signBit:signBit
    ^ ((self basicNew:(numBits // 8)) exponentSize:exponentSize) 
            setFraction:anInteger exponent:exp signBit:signBit

    "
     self size:256 exponentSize:19 fromInteger:1
     self size:256 exponentSize:19 fromInteger:2
    "
!

size:numBits exponentSize:exponentSize fromFloat:aFloat
    ^ ((self basicNew:(numBits // 8)) exponentSize:exponentSize) setValueFromFloat:aFloat

    "
     self size:256 exponentSize:19 fromFloat:1.0
     self size:256 exponentSize:19 fromFloat:2.0
    "
!

size:numBits exponentSize:exponentSize fromInteger:anInteger
    ^ ((self basicNew:(numBits // 8)) exponentSize:exponentSize) setValueFromInteger:anInteger

    "
     self size:256 exponentSize:19 fromInteger:1
     self size:256 exponentSize:19 fromInteger:2
    "
! !

!IEEEFloat class methodsFor:'coercing & converting'!

coerce:aNumber
    "convert the argument aNumber into an instance of the receiver (class) and return it."

    ^ aNumber asIEEEFloat
! !

!IEEEFloat class methodsFor:'constants'!

unity
    ^ 1.0
!

zero
    ^ 0.0
! !

!IEEEFloat methodsFor:'accessing'!

exponentBits
    "return the bits of my exponent.
     These might be biased.
     My bytes are stored LSB first; thus we have to fetch the high exponent bytes."

    |nBytes nExpBytes e|

    nBytes := self basicSize.
    nExpBytes := (exponentSize + 1 + 7) // 8.
    "/ seee eeee eeee mmmm mmmm ....
    "/ BAD: assumes LSB
    e := (self basicAt:nBytes) bitAnd:16r7F.
    2 to:nExpBytes do:[:eI |
         e := (e bitShift:8) bitOr:(self basicAt:nBytes + 1 - eI)
    ].
    "/ 0eee eeee eeee mmmm
    e := e rightShift:(nExpBytes * 8) - 1 "sign" - exponentSize.
    "/ 0000 0eee eeee eeee 
    ^ e

    "
     1.0 asShortFloat asIEEEFloat
     (1 to:4) conform:[:i | (1.0 asShortFloat basicAt:i) = (1.0 asShortFloat asIEEEFloat basicAt:i)]
     1.0 asShortFloat exponentBits 
     1.0 asShortFloat asIEEEFloat exponentBits
     -1.0 asShortFloat exponentBits  
     -1.0 asShortFloat asIEEEFloat exponentBits   

     (self size:32 exponentSize:8) inspect.
     (self size:64 exponentSize:11) inspect.
     (self size:128 exponentSize:16) inspect.
     (self size:256 exponentSize:16) inspect.
    "

    "construct a soft-shortFloat

     |f sf|
     sf := 1.0 asShortFloat.
     f := self size:(sf basicSize * 8) exponentSize:(sf numBitsInExponent).
     1 to:sf basicSize do:[:i | f basicAt:i put:(sf basicAt:i)].
     f inspect
    "

    "Modified: 12.2.1997 / 16:45:27 / cg"
!

exponentSize
    ^ exponentSize
!

exponentSize:something
    exponentSize := something.
!

mantissaBits
    "return the bits of my mantissa (incl. the hidden bit).
     My bytes are stored LSB first."

    |bits hiddenBit mask c|

    bits := self digitBytes asIntegerMSB:false.
    bits == 0 ifTrue:[^ 0].

    hiddenBit := (1 bitShift:self numBitsInMantissa).
    mask := hiddenBit-1.
    c := bits bitAnd:mask.
    "/ TODO: care for subnormals (exponent = 0)!!
    ^ c bitOr:hiddenBit.

    " 
     (self size:32 exponentSize:8) inspect.
    "

   "construct a soft-shortFloat

    |f sf|
    sf := 1.0 asShortFloat.
    f := self size:(sf basicSize * 8) exponentSize:(sf numBitsInExponent).
    1 to:sf basicSize do:[:i | f basicAt:i put:(sf basicAt:i)].
    f inspect
   "

    "Modified: 12.2.1997 / 16:45:27 / cg"
!

setFraction:fraction exponent:exponent signBit:signBit
    |bits|

    bits := (fraction 
            bitOr:(exponent bitShift:self numBitsInMantissa))
            bitOr:(signBit bitShift:(self basicSize * 8) - 1). 
    1 to:self basicSize do:[:i | self basicAt:i put:(bits digitAt:i)].

    "
     (self size:64 exponentSize:11) setFraction:0 exponent:1024 signBit:0
    "
!

setValueFromInteger:intValue
    "/ how many bits are there, in this int
    |absValue myNumBits numBitsInNumber shift|

    absValue := intValue abs. 
    numBitsInNumber := absValue highBit.
    myNumBits := (self basicSize * 8) - 1 "sign" - exponentSize.
    shift := myNumBits - numBitsInNumber. 
    numBitsInNumber > myNumBits ifTrue:[
        self halt.
    ] ifFalse:[
        "/ number:
        "/    1xxxxxxx...xxxxx
        "/ myRep:
        "/    seee...eeexxxxxxxxxx
        absValue digitLength to:1 by:-1 do:[:byteIndex |
            
        ].
    ].

    "/ cut off some bits

    "Float numBitsInExponent
     self size:16 exponentSize:4 fromInteger:1
     self size:32 exponentSize:11 fromInteger:1

     self size:256 exponentSize:19 fromInteger:1
     self size:256 exponentSize:19 fromInteger:2
    "
! !

!IEEEFloat methodsFor:'arithmetic'!

* aNumber
    "/ thisContext isRecursive ifTrue:[self halt].
    ^ aNumber productFromIEEEFloat:self
!

+ aNumber
    "/ thisContext isRecursive ifTrue:[self halt].
    ^ aNumber sumFromIEEEFloat:self

    "
     (-2.0 asIEEEFloat)
     (2.0 asIEEEFloat)

     self assert:((4.0 asIEEEFloat) + (2.0 asIEEEFloat)) = (6.0 asIEEEFloat)   
     self assert:((4.0 asIEEEFloat) + (-2.0 asIEEEFloat)) = (2.0 asIEEEFloat)  
     self assert:((-4.0 asIEEEFloat) + (-2.0 asIEEEFloat)) = (-6.0 asIEEEFloat)  
     self assert:((-4.0 asIEEEFloat) + (2.0 asIEEEFloat)) = (-2.0 asIEEEFloat)  

     self assert:((2.0 asIEEEFloat) + (4.0 asIEEEFloat)) = (6.0 asIEEEFloat)   
     self assert:((2.0 asIEEEFloat) + (-4.0 asIEEEFloat)) = (-2.0 asIEEEFloat)  
     self assert:((-2.0 asIEEEFloat) + (-4.0 asIEEEFloat)) = (-6.0 asIEEEFloat)  
     self assert:((-2.0 asIEEEFloat) + (4.0 asIEEEFloat)) = (2.0 asIEEEFloat)  

     self assert:((IEEEFloat fromFloat:0.1) + (IEEEFloat fromFloat:-0.1)) = (IEEEFloat fromFloat:0.0)  
     self assert:((IEEEFloat fromFloat:-0.1) + (IEEEFloat fromFloat:0.1)) = (IEEEFloat fromFloat:0.0)  
    "
!

- aNumber
    "/ thisContext isRecursive ifTrue:[self halt].
    ^ aNumber differenceFromIEEEFloat:self

    "
     self assert:((4.0 asIEEEFloat) - (2.0 asIEEEFloat)) = (2.0 asIEEEFloat)   
     self assert:((4.0 asIEEEFloat) - (-2.0 asIEEEFloat)) = (6.0 asIEEEFloat)  
     self assert:((-4.0 asIEEEFloat) - (-2.0 asIEEEFloat)) = (-2.0 asIEEEFloat)  
     self assert:((-4.0 asIEEEFloat) - (2.0 asIEEEFloat)) = (-6.0 asIEEEFloat)  

     self assert:((2.0 asIEEEFloat) - (4.0 asIEEEFloat)) = (-2.0 asIEEEFloat)   
     self assert:((2.0 asIEEEFloat) - (-4.0 asIEEEFloat)) = (6.0 asIEEEFloat)  
     self assert:((-2.0 asIEEEFloat) - (-4.0 asIEEEFloat)) = (2.0 asIEEEFloat)  
     self assert:((-2.0 asIEEEFloat) - (4.0 asIEEEFloat)) = (-6.0 asIEEEFloat)  

     self assert:((IEEEFloat fromFloat:0.1) - (IEEEFloat fromFloat:0.1)) = (IEEEFloat fromFloat:0.0)  
     self assert:((IEEEFloat fromFloat:0.1) - (IEEEFloat fromFloat:-0.1)) = (IEEEFloat fromFloat:0.2)  
     self assert:((IEEEFloat fromFloat:-0.1) - (IEEEFloat fromFloat:-0.1)) = (IEEEFloat fromFloat:0.0)  
    "
!

/ aNumber
    "/ thisContext isRecursive ifTrue:[self halt].
    ^ aNumber quotientFromIEEEFloat:self
!

negated
    |f highByte|

    f := self copy.
    highByte := self basicAt:(self basicSize).
    f basicAt:(self basicSize) put:(highByte bitXor:16r80).
    ^ f.

    "
     self assert:(IEEEFloat fromFloat:2.0) negated = (IEEEFloat fromFloat:-2.0) 
     self assert:(IEEEFloat fromFloat:2.0) negated negated = (IEEEFloat fromFloat:2.0) 
     self assert:(IEEEFloat fromFloat:-2.0) negated = (IEEEFloat fromFloat:2.0) 
    "
! !

!IEEEFloat methodsFor:'coercing & converting'!

generality
    ^ 97   "/ between OctaFloat and LargeFloat

    "
     1 asShortFloat generality 70
     1 asFloat generality      80
     1 asLongFloat generality  90
     1 asQuadFloat generality
     1 asQDouble generality    95
     1 asLargeFloat generality 100
    "
! !

!IEEEFloat methodsFor:'comparing'!

< aNumber
    "/ thisContext isRecursive ifTrue:[self halt].
    ^ aNumber lessFromIEEEFloat:self
!

= aNumber
    ^ aNumber equalFromIEEEFloat:self
! !

!IEEEFloat methodsFor:'double dispatching'!

differenceFromIEEEFloat:anIEEEFloat
    "/ anIEEEFloat - self
    ^ self negated sumFromIEEEFloat:anIEEEFloat
!

equalFromIEEEFloat:anIEEEFloat
    |m1 m2 nM1 nM2 e1 e2|

    anIEEEFloat basicSize == self basicSize ifTrue:[
        anIEEEFloat exponentSize == exponentSize ifTrue:[
            1 to:self basicSize do:[:i |
                (self basicAt:i) = (anIEEEFloat basicAt:i) ifFalse:[^ false].
            ].
            ^ true.
        ].
    ].

    "/ more complicated compare

    self negative = anIEEEFloat negative ifFalse:[^ false].

    "/ bring to same exponent, add m1 + m2, normalize

    m1 := anIEEEFloat mantissaBits. 
    m2 := self mantissaBits.        

    nM1 := anIEEEFloat numBitsInMantissa.
    nM2 := self numBitsInMantissa.
    nM1 ~= nM2 ifTrue:[
        nM1 > nM2 ifTrue:[
            m2 := m2 bitShift:(nM1-nM2).
        ] ifFalse:[
            m1 := m1 bitShift:(nM2-nM1).
        ].
    ].
    m1 = m2 ifFalse:[^ false].

    e1 := anIEEEFloat exponentBits - anIEEEFloat eBias.
    e2 := self exponentBits - self eBias.
    e1 = e2 ifFalse:[^ false].

    ^ true.
!

lessFromIEEEFloat:anIEEEFloat
    ||

    anIEEEFloat basicSize == self basicSize ifTrue:[
    ].
    self halt.
!

productFromIEEEFloat:anIEEEFloat
    "/ anIEEEFloat * self

    |m1 m2 e1 e2 m e signBit hi shift nM1 nM2 nM|

    "/ m1 * m2 / e1 + e2

    m1 := anIEEEFloat mantissaBits. 
    m2 := self mantissaBits.        

    nM1 := anIEEEFloat numBitsInMantissa.   
    nM2 := nM := self numBitsInMantissa.
    nM1 ~= nM2 ifTrue:[
        nM1 > nM2 ifTrue:[
            m2 := m2 bitShift:(nM1-nM2).
            nM := nM1
        ] ifFalse:[
            m1 := m1 bitShift:(nM2-nM1).
            nM := nM2
        ].
    ].

    e1 := anIEEEFloat exponentBits.
    e2 := self exponentBits.
    e := (e1 - anIEEEFloat eBias) + (e2 - self eBias).

    "/ 1xxxx....xxxx * 1xxxx....xxxx
    m := m1 * m2.
    "/ now have nB bits too many
    e := e - nM.

    signBit := (self negative == anIEEEFloat negative) ifTrue:[0] ifFalse:[1].
    m == 0 ifTrue:[self halt].

    hi := m highBit.    
    shift := (nM + 1) - hi. 
    m := m bitShift:shift.
    e := e - shift.
    m := m bitAnd:(1 bitShift:nM)-1. 

    e := e + self eBias.
    ^ self class 
        size:(self basicSize*8) exponentSize:exponentSize 
        fraction:m exponent:e signBit:signBit.

    "
     self assert:((IEEEFloat fromFloat:1.0) * (IEEEFloat fromFloat:1.0)) = (IEEEFloat fromFloat:1.0) 
     self assert:((IEEEFloat fromFloat:1.0) * (IEEEFloat fromFloat:1.0)) exponent = 1.0 exponent
     self assert:((IEEEFloat fromFloat:1.0) * (IEEEFloat fromFloat:1.0)) mantissaBits = 1.0 mantissaBits

     self assert:((4.0 asIEEEFloat) * (2.0 asIEEEFloat)) = (8.0 asIEEEFloat)  
     self assert:((-4.0 asIEEEFloat) * (2.0 asIEEEFloat)) = (-8.0 asIEEEFloat)  
     self assert:((4.0 asIEEEFloat) * (-2.0 asIEEEFloat)) = (-8.0 asIEEEFloat)  
     self assert:((-4.0 asIEEEFloat) * (-2.0 asIEEEFloat)) = (8.0 asIEEEFloat)  

     self assert:((IEEEFloat fromFloat:100.0) * (IEEEFloat fromFloat:100.0)) = (IEEEFloat fromFloat:10000.0) 
     self assert:((IEEEFloat fromFloat:1e-20) * (IEEEFloat fromFloat:1e20)) = (IEEEFloat fromFloat:(1e-20 * 1e20)) 
    "
!

quotientFromIEEEFloat:anIEEEFloat
    "/ anIEEEFloat / self

    |m1 m2 e1 e2 m e signBit nM1 nM2 nM hi shift|

    "/ m1 / m2 / e1 - e2

    m1 := anIEEEFloat mantissaBits. 
    m2 := self mantissaBits.        

    nM1 := anIEEEFloat numBitsInMantissa.
    nM2 := nM := self numBitsInMantissa.
    nM1 ~= nM2 ifTrue:[
        nM1 > nM2 ifTrue:[
            m2 := m2 bitShift:(nM1-nM2).
            nM := nM1
        ] ifFalse:[
            m1 := m1 bitShift:(nM2-nM1).
            nM := nM2
        ].
    ].

    e1 := anIEEEFloat exponentBits.
    e2 := self exponentBits.
    e := (e1 - anIEEEFloat eBias) - (e2 - self eBias).

    "/ 1xxxx....xxxx / 1xxxx....xxxx
    m := (m1 bitShift:nM) // m2.
    "/ rest := m1 \\ m2.

    signBit := (self negative == anIEEEFloat negative) ifTrue:[0] ifFalse:[1].
    m == 0 ifTrue:[self halt].

    hi := m highBit.    
    shift := (nM + 1) - hi. 
    m := m bitShift:shift.
    e := e - shift.
    m := m bitAnd:(1 bitShift:nM)-1. 

    e := e + self eBias.
    ^ self class 
        size:(self basicSize*8) exponentSize:exponentSize 
        fraction:m exponent:e signBit:signBit.

    "
     self assert:((IEEEFloat fromFloat:1.0) / (IEEEFloat fromFloat:1.0)) = (IEEEFloat fromFloat:1.0) 
     self assert:((IEEEFloat fromFloat:1.0) / (IEEEFloat fromFloat:1.0)) exponent = 1.0 exponent
     self assert:((IEEEFloat fromFloat:1.0) / (IEEEFloat fromFloat:1.0)) mantissaBits = 1.0 mantissaBits

     self assert:((4.0 asIEEEFloat) / (2.0 asIEEEFloat)) = (2.0 asIEEEFloat)  
     self assert:((-4.0 asIEEEFloat) / (2.0 asIEEEFloat)) = (-2.0 asIEEEFloat)  
     self assert:((4.0 asIEEEFloat) / (-2.0 asIEEEFloat)) = (-2.0 asIEEEFloat)  
     self assert:((-4.0 asIEEEFloat) / (-2.0 asIEEEFloat)) = (2.0 asIEEEFloat)  

     self assert:((IEEEFloat fromFloat:100.0) / (IEEEFloat fromFloat:2.0)) = (IEEEFloat fromFloat:50.0) 
     self assert:((IEEEFloat fromFloat:1e-20) / (IEEEFloat fromFloat:1e20)) = (IEEEFloat fromFloat:(1e-20 / 1e20)) 
    "
!

sumFromIEEEFloat:anIEEEFloat
    "/ anIEEEFloat + self

    |m1 m2 e1 e2 m e signBit hi shift nM1 nM2 nM|

    "/ bring to same exponent, add m1 + m2, normalize

    m1 := anIEEEFloat mantissaBits. m1 == 0 ifTrue:[^ self].
    m2 := self mantissaBits.        m2 == 0 ifTrue:[^ anIEEEFloat].

    nM1 := anIEEEFloat numBitsInMantissa.
    nM2 := nM := self numBitsInMantissa.
    nM1 ~= nM2 ifTrue:[
        nM1 > nM2 ifTrue:[
            m2 := m2 bitShift:(nM1-nM2).
            nM := nM1
        ] ifFalse:[
            m1 := m1 bitShift:(nM2-nM1).
            nM := nM2
        ].
    ].

    e1 := anIEEEFloat exponentBits - anIEEEFloat eBias.
    e2 := self exponentBits - self eBias.
    e := e1.
    e1 = e2 ifFalse:[
        e1 > e2 ifTrue:[
            m2 := m2 rightShift:(e1 - e2).
            e := e1.
        ] ifFalse:[
            "/ e2 > e1
            m1 := m1 rightShift:(e2 - e1).
            e := e2.
        ]
    ].

    signBit := 0.
    "/ ok, add the mantissae
    self negative ifTrue:[
        anIEEEFloat negative ifTrue:[
            "/ - (anIEEEFloat + self)
            m := m1 + m2.
            signBit := 1.
        ] ifFalse:[
            "/ anIEEEFloat - self 
            m := m1 - m2
        ].
    ] ifFalse:[
        anIEEEFloat negative ifTrue:[
            "/ self - anIEEEFloat
            m := m2 - m1.
        ] ifFalse:[
            "/ (anIEEEFloat + self)
            m := m1 + m2.
        ].
    ].
    m == 0 ifTrue:[^ self class zero].
    m < 0 ifTrue:[
        m := m negated.
        signBit := 1-signBit
    ].

    hi := m highBit.    
    shift := (nM + 1) - hi. 
    m := m bitShift:shift.
    e := e - shift.
    m := m bitAnd:(1 bitShift:nM)-1. 

    ^ self class 
        size:(self basicSize*8) exponentSize:exponentSize 
        fraction:m exponent:(e + self eBias) signBit:signBit.

    "
     seee eeee  eeee mmmm  mmmm mmmm  mmmm mmmm  mmmm mmmm  mmmm mmmm  mmmm mmmm  mmmm mmmm 
     0100 0000  0000 0000  0000 ......                                                       2.0
     0011 1111  1111 0000  0000 ......                                                       1.0
     0400 0000  0001 0000  0000 ......                                                       4.0

     self assert:(IEEEFloat fromFloat:1.0) exponent = 1.0 exponent.  
     self assert:(IEEEFloat fromFloat:1.0) mantissaBits = 1.0 mantissaBits.
     self assert:(IEEEFloat fromFloat:1.0) eBias = 1.0 eBias.

     self assert:(IEEEFloat fromFloat:2.0) exponent = 2.0 exponent.  
     self assert:(IEEEFloat fromFloat:2.0) mantissaBits = 2.0 mantissaBits.
     self assert:(IEEEFloat fromFloat:2.0) eBias = 2.0 eBias.

     self assert:(IEEEFloat fromFloat:4.0) exponent = 4.0 exponent.  
     self assert:(IEEEFloat fromFloat:4.0) mantissaBits = 4.0 mantissaBits.
     self assert:(IEEEFloat fromFloat:4.0) eBias = 4.0 eBias.

     self assert:(IEEEFloat fromFloat:3.0) exponent = 3.0 exponent.  
     self assert:(IEEEFloat fromFloat:3.0) mantissaBits = 3.0 mantissaBits.
     self assert:(IEEEFloat fromFloat:3.0) eBias = 3.0 eBias.

     self assert:(IEEEFloat fromFloat:10.0) exponent = 10.0 exponent.  
     self assert:(IEEEFloat fromFloat:10.0) mantissaBits = 10.0 mantissaBits.
     self assert:(IEEEFloat fromFloat:10.0) eBias = 10.0 eBias.

     self assert:(IEEEFloat fromFloat:1e-3) exponent = 1e-3 exponent.  
     self assert:(IEEEFloat fromFloat:1e-3) mantissaBits = 1e-3 mantissaBits.
     self assert:(IEEEFloat fromFloat:1e-3) eBias = 1e-3 eBias.

     (IEEEFloat fromFloat:1.0) + (IEEEFloat fromFloat:1.0) = (IEEEFloat fromFloat:2.0) 
     ((IEEEFloat fromFloat:1.0) + (IEEEFloat fromFloat:1.0)) exponent = 2.0 exponent
     ((IEEEFloat fromFloat:1.0) + (IEEEFloat fromFloat:1.0)) mantissaBits = 2.0 mantissaBits

     self assert:((-4.0 asIEEEFloat) + (2.0 asIEEEFloat)) = (-2.0 asIEEEFloat)  

     (IEEEFloat fromFloat:2.0)
     self assert:((IEEEFloat fromFloat:2.0) + (IEEEFloat fromFloat:2.0)) = (IEEEFloat fromFloat:4.0) 
     self assert:((IEEEFloat fromFloat:4.0) + (IEEEFloat fromFloat:4.0)) = (IEEEFloat fromFloat:8.0) 
     self assert:((IEEEFloat fromFloat:0.0) + (IEEEFloat fromFloat:1.0)) = (IEEEFloat fromFloat:1.0) 
     self assert:((IEEEFloat fromFloat:1.0) + (IEEEFloat fromFloat:0.0)) = (IEEEFloat fromFloat:1.0) 
     self assert:((IEEEFloat fromFloat:1.0) + (IEEEFloat fromFloat:2.0)) = (IEEEFloat fromFloat:3.0) 
     self assert:((IEEEFloat fromFloat:100.0) + (IEEEFloat fromFloat:2.0)) = (IEEEFloat fromFloat:102.0) 
     self assert:((IEEEFloat fromFloat:2.0) + (IEEEFloat fromFloat:100.0)) = (IEEEFloat fromFloat:102.0) 

     self assert:((IEEEFloat fromFloat:4.0) + (IEEEFloat fromFloat:-2.0)) = (IEEEFloat fromFloat:2.0)  
     self assert:((IEEEFloat fromFloat:8.0) + (IEEEFloat fromFloat:-2.0)) = (IEEEFloat fromFloat:6.0)  
     self assert:((IEEEFloat fromFloat:100.0) + (IEEEFloat fromFloat:-2.0)) = (IEEEFloat fromFloat:98.0)  
     self assert:((IEEEFloat fromFloat:-2.0) + (IEEEFloat fromFloat:100.0)) = (IEEEFloat fromFloat:98.0)  
     self assert:((IEEEFloat fromFloat:-2.0) + (IEEEFloat fromFloat:-100.0)) = (IEEEFloat fromFloat:-102.0)  

     self assert:((IEEEFloat fromFloat:-0.1) + (IEEEFloat fromFloat:0.1)) = (IEEEFloat fromFloat:0.0)  
    "
! !

!IEEEFloat methodsFor:'inspecting'!

inspectorExtraAttributes
    ^ #()
! !

!IEEEFloat methodsFor:'printing & storing'!

displayOn:aStream
aStream nextPutAll:'IEEEFloat'.
^ self.
!

displayString
^ 'IEEEFloat'.

!

printOn:aStream
aStream nextPutAll:'IEEEFloat'.
^ self.
    thisContext isRecursive ifTrue:[
        aStream nextPutAll:'IEEEFloat (error while printing)'.
        ^ self.
    ].

    PrintfScanf printf:'%g' on:aStream argument:self.
!

printOn:aStream base:b
aStream nextPutAll:'IEEEFloat'.
!

printOn:aStream base:b showRadix:showRadix
aStream nextPutAll:'IEEEFloat'.
!

printOn:aStream thousandsSeparator:thousandsSeparator
aStream nextPutAll:'IEEEFloat'.
!

printString
^ 'IEEEFloat'.
!

printStringFormat:formatString
^ 'IEEEFloat'.
!

printStringRadix:base
^ 'IEEEFloat'.
!

printStringRadix:base showRadix:showRadixBoolean
^ 'IEEEFloat'.
!

printStringScientific
^ 'IEEEFloat'.
!

printfPrintString:formatString
^ 'IEEEFloat'.
! !

!IEEEFloat methodsFor:'queries'!

eBias
    "Answer the exponent's bias;
     that is the offset of the zero exponent when stored"

    ^ (1 bitShift:exponentSize-1) - 1

    "
     1.0 numBitsInExponent 11
     1.0 eBias             1023
     1.0 emin              -1022
     1.0 emax              1023
     1.0 fmin              2.2250738585072E-308
     1.0 fmax              1.79769313486232E+308
    "
!

numBitsInExponent
    ^ exponentSize

    "
     1.0 numBitsInExponent 11
     1.0 numBitsInMantissa 52
     1.0 precision         53
     1.0 decimalPrecision  16
     1.0 eBias             1023
     1.0 emin              -1022
     1.0 emax              1023
     1.0 fmin              2.2250738585072E-308
     1.0 fmax              1.79769313486232E+308
    "
!

numBitsInMantissa
    "answer the number of bits in the mantissa (the significant).
     This is an IEEE float,
        where n*8 - 1 - exponentSize (n=nr of bytes)
     bits are available in the mantissa (the hidden bit is not counted here):
        s ee...ee mmm...mmm
    "

    ^ (self basicSize * 8) - exponentSize - 1 "/ sign
! !

!IEEEFloat methodsFor:'testing'!

isFinite
    "return true, if the receiver is a finite float (not NaN or +/-INF)"

   ^ self exponentBits ~= ((1 bitShift:self numBitsInExponent)-1)

    "Modified: 12.2.1997 / 16:45:27 / cg"
!

isInfinite
    "return true, if the receiver is an infinite float (+/-INF)"

   ^ (self exponentBits = ((1 bitShift:self numBitsInExponent)-1))
   and:[ self mantissaBits = 0 ]

    "Modified: 12.2.1997 / 16:45:27 / cg"


!

isNaN
    "return true, if the receiver is an invalid float (NaN - not a number)"

   ^ (self exponentBits = ((1 bitShift:self numBitsInExponent)-1))
   and:[ self mantissaBits ~= 0 ]

    "Modified: 12.2.1997 / 16:45:27 / cg"
!

isZero
    "return true, if the receiver is zero"

    1 to:self size do:[:i | 
        (self basicAt:i) ~~ 0 ifTrue:[^ false].
    ].
    ^ true

!

negative
    "return true if the receiver is less than zero."

    ^ (self basicAt:self basicSize) bitTest:16r80
!

positive
    "return true if the receiver is greater or equal to zero."

    ^ ((self basicAt:self basicSize) bitTest:16r80) not
! !

!IEEEFloat class methodsFor:'documentation'!

version_CVS
    ^ '$Header$'
! !

